
package wesen.robot;

import java.lang.*;
import java.awt.*;
import java.util.*;
import java.awt.geom.*;
import javax.swing.*;
import java.awt.Component;
import java.awt.image.BufferedImage;
import java.io.*;
import com.symantec.itools.javax.swing.JMaskedTextField;
import java.net.URL;
import wesen.simulator.*;
import wesen.util.*;

/**
 * Diese Klasse implementiert das AnimationVector Interface und verwaltet alle
 * Objekte der Simulationswelt. Des Simulator kommuniziert nur dem Objekt
 * dieser Klasse, d.h. zu den Robotern usw. hat dieser keine Referenz.
 * 
 * @author Wesen-Team
 * @version 1.0
 */
public class ObjectVector extends Vector implements AnimationVector {

  /**
   * Ueberschreiben der serialisierungs-ID, damit werden InvalidClassException ausgeschlossen.
   * Besser waere alle nicht fuer die Serialsierung benutzten Variablen als transient zu
   * deklarieren, dann wuerde bei einer Aenderung der nicht relevanten Variablen dieses
   * Problem nicht auftreten.
   */
  static final long             serialVersionUID = 5229886082550730620L;

  /**
   * Verzeichnis der Roboter und Aktoren Dateien, diese beschreiben die Eigenschaften
   * eines Roboters bzw. Aktoren
   */
  final private static String   DIR_OBJECTS = "RobotData";

  /**
   * Filenamensuffix fuer die Beschreibung der Roboter
   */
  final private static String   SUFFIX_XML = ".xml";


  /**
   * Datei welche in dieser Datei aufgelistet sind werden moeglicherweise 
   * aus der JAR-Datei gelesen und in das Benutzerverzeichnis gespeichert.
   */ 
  final private static String  DIR_LIST = "dirlist.txt";


  /**
   * Filenamensuffix fuer die Beschreibung der Aktoren
   */
  final private static String   SUFFIX_PROP = ".properties";
  private static NewObservable  observable = new NewObservable();

  /**
   * -1 = kein Objekt markiert; >0 Der Index der markierten Objektes
   */
  private static int            oldMarkId = -1;

  /**
   * Enthaelt moeglicherweise eine Objekt. Diese Objekt kann Eingefuegt werden.
   * @supplierCardinality 0..1
   * @supplierQualifier besetzt Clipboard
   */
  private AnimationObject       clipboardObject;

  /**
   * @clientCardinality 1
   * @supplierCardinality *
   * @supplierRole hat
   */

  /*#  private AnimationObject lnkAnimationObject; */

  /**
   * Der Konstruktor wird nicht direkt angesprochen
   */
  public ObjectVector() {
    super();
  }

  /**
   * Hinzufuegen eines Objektes.
   */
  public void addObject(int x, int y, String fileName) {
    AnimationObject a;

    if (fileName.endsWith(SUFFIX_XML)) {

      // Es ist ein Roboter
      a = Roboter.getNewInstanceXML(x, y, fileName, getDirOfRobotData());
    } else {

      // Es ist ein Aktor
      a = new Actor(x, y, fileName, getDirOfRobotData());
    }   // else
 
    // Die Id muss zuerst erstellt werden erst danach kann das Objekt hinzugefuegt werden
    if (a != null) {
      setFreeIDNumber(a);
      add(a);
      observable.notifyObservers(this);
      if (a instanceof Roboter) {
        // Es muss eine neue Kopie fuer das Reset bereitgestellt werden
        ((Roboter) a).setMemento((Roboter) a);
      }   // if
    }  // if
  }   // addObject
 
  /**
   * Erster Simulationscyklus
   */
  public synchronized  void initRoboter() {
    int size = size();

    for (int i = 0; i < size; i++) {
      if (elementAt(i) instanceof Roboter) {
        ((Roboter) elementAt(i)).initRoboter();
      }   // if
     }    // for
   } 

  /**
   * Method declaration
   *
   *
   * @param observer
   *
   * @see
   */
  public void registerObserver(Observer observer) {
    observable.addObserver(observer);
  } 

  /**
   * Objekte in Ausgangspostion bringen.
   */
  public void resetPosition() {
    int size = size();

    for (int i = 0; i < size; i++) {
      if (elementAt(i) instanceof Roboter) {
        // ((Roboter) elementAt(i)).resetPosition();
        setElementAt(((Roboter) elementAt(i)).getMemento(), i);
      }   // if
     }    // for
   }      // resetRoboter
 
  /**
   * Method declaration
   *
   *
   * @return
   *
   * @see
   */
  public int countNumberOfRoboters() {
    int size = size();
    int countRoboter = 0;

    for (int i = 0; i < size; i++) {
      if (elementAt(i) instanceof Roboter) {
        countRoboter++;
      }   // if
     }    // for
 
    return countRoboter;
  } 

  /**
   * Method declaration
   *
   *
   * @param iX
   * @param iY
   * @param umgebung
   *
   * @see
   */
  public void startBeanEditor(int iX, int iY, Umgebung umgebung) {
    AnimationObject ao = getObjectAtPoint(iX, iY);

    ao.registerBeanEditor(umgebung);
  } 

  /**
   * Hier werden die Roboter neu platziert. In naechsten Bild werden
   * diese so angzeigt.
   */
  public synchronized void moveRoboter() {
    int size = size();

    for (int i = 0; i < size; i++) {

      // Hier werden die entsprechenden Methodenaufrufe fuer die Verschiebung
      // der Roboter aufgerufen
      if (elementAt(i) instanceof Roboter) {
        ((Roboter) elementAt(i)).move(this, i);

        // TODO: mike
      }   // if
     }    // for
 
    handleCollision();
  }   // moveRoboter
 
  /**
   * Die folgenden Methode ist nur als Beispiel gedacht und wird in userer
   * Simulation wohl so nicht gebraucht.
   */
  private void handleCollision() {
    AnimationObject r1;
    AnimationObject r2;
    int             size = size();

    for (int i = 0; i < size; i++) {
      for (int j = 0; j < size; j++) {
        if (i != j) {
          r1 = (AnimationObject) elementAt(i);
          r2 = (AnimationObject) elementAt(j);

          if (collide(r1, r2)) {
            r1.hit(r2);
            r2.hit(r1);
          }   // if
         }    // if
       }      // for
     }        // for
   }          // handelCollision
 
  /**
   * Sucht ob der gegebene Punkt durch ein Objekt (Roboter bzw. Aktor) besetzt ist.
   */
  public int getObjectAtPointID(int iX, int iY) {
    int size = size();
    int i = 0;

    while (i < size &&!((AnimationObject) elementAt(i)).pointIsPart(iX, iY)) {
      i++;
    } 

    return ((i < size) ? i : -1);
  } 

  /**
   * Method declaration
   *
   *
   * @param iX
   * @param iY
   *
   * @return
   *
   * @see
   */
  public AnimationObject getObjectAtPoint(int iX, int iY) {
    int i = getObjectAtPointID(iX, iY);

    return (AnimationObject) ((i == -1) ? null : elementAt(i));
  } 

  /**
   * Method declaration
   *
   *
   * @param iX
   * @param iY
   *
   * @return
   *
   * @see
   */
  public String getToolTipOfObjectAtPoint(int iX, int iY) {
    String  strToolTip = null;
    int     i;

    if ((i = getObjectAtPointID(iX, iY)) != -1) {
      strToolTip = ((AnimationObject) elementAt(i)).getToolTipText();
    } 

    return strToolTip;
  } 

  /**
   * Setzt das Objekt (Aktor bzw. Roboter) entsprechend seiner x und y Position
   */
  public void setPos(int workId, int iX, int iY) {
    ((AnimationObject) elementAt(workId)).setPos(iX, iY);
  } 

  /**
   * Objekt ins Clipboard setzen.
   */
  public boolean setClipboard() {
    int     workId = getMark();
    boolean hasSet = false;

    try {
      if (workId >= 0) {
        clipboardObject = ((AnimationObject) elementAt(workId)).cloneObject();
        hasSet = true;
      }   // if
     } catch (Exception e) {
      e.printStackTrace();
    } 

    return hasSet;
  } 

  /**
   * Winkel fuer das Objekt setzen.
   */
  public void setAlphaGrad(double dGrad, int markId) {
    ((Roboter) elementAt(markId)).setGrad(dGrad);
  } 

  /**
   * Markierung fuer ein Objekt setzen bzw. entfernen.
   * Gleichzeitig kann nur ein Objekt markiert sein.
   */
  public void setMark(Simulator simulator, int workId, boolean isMarkOn) {
    if (workId != oldMarkId && oldMarkId >= 0) {
      ((AnimationObject) elementAt(oldMarkId)).setMark(false);
      simulator.setAlpha(false, 0);

      oldMarkId = -1;
    }   // if
 
    if (workId >= 0) {
      oldMarkId = (isMarkOn) ? workId : -1;

      ((AnimationObject) elementAt(workId)).setMark(isMarkOn);

      if (isMarkOn && elementAt(workId) instanceof Roboter) {
        simulator.setAlpha(true, ((Roboter) elementAt(workId)).getAlpha());
      } else {
        simulator.setAlpha(false, 0);
      }   // else
     }    // if
 
    simulator.setCopy(isMarkOn);
  }   // setMark
 
  /**
   * Rueckgabe des markierten Objektes.
   */
  public int getMark() {
    return oldMarkId;
  } 

  /**
   * Rueckgabe der Durchmessers des entsprechenden Objektes.
   */
  public int getObjectDiameter(int workId) {
    return ((AnimationObject) elementAt(workId)).getDiameter();
  } 

  /**
   * Speichern der Welt. Jedes Objekt speichert sich selbst und seine abhaengige Objekte.
   */
  public void saveWorld(ObjectOutputStream oos) throws IOException {
    int size = size();

    for (int i = 0; i < size; i++) {
      ((AnimationObject) elementAt(i)).saveAnimationObject(oos);
    }   // for
   } 

  /**
   * Einlesen der serialisierten Objekte.
   */
  public void readWorld(ObjectInputStream ois) throws IOException {
    try {
      while (true) {
        Class           cl = (Class) ois.readObject();

        // Erstellt eine Instanc des Objektes
        AnimationObject ao = (AnimationObject) cl.newInstance();

        ao = ao.loadAnimationObject(ois, getDirOfRobotData());

        if (ao != null) {
          setFreeIDNumber(ao);
          add(ao);
          if (ao instanceof Roboter) {
            // Es muss eine neue Kopie fuer das Reset bereitgestellt werden
           ((Roboter) ao).setMemento((Roboter) ao);
          }   // if
        } else {

          // Das Objekt konnte nicht in den Vector aufgenommen werden.
          MessageHandler.showTextFieldMessage("readworld.error1", 
                                              MessageHandler.ERROR_COLOR);
        }   // else
       }    // while
     } catch (EOFException eof) {}
    catch (ClassNotFoundException cnfe) {
      cnfe.printStackTrace();
    } catch (InstantiationException ie) {
      ie.printStackTrace();
    } catch (IllegalAccessException iae) {
      iae.printStackTrace();
    }       // catch
     finally {
      observable.notifyObservers(this);
    } 
  } 

  /**
   * Sucht den Punktes der am weitesten von der Koordinate (0,0)
   * durch ein Objekte belegt ist.
   */
  public Dimension getMaxOccupiedYX() {
    double  x = 0;
    double  y = 0;
    int     size = size();

    for (int i = 0; i < size; i++) {
      AnimationObject ao = (AnimationObject) elementAt(i);

      x = Math.max(ao.getX() + ao.getDiameter() / 2, x);
      y = Math.max(ao.getY() + ao.getDiameter() / 2, y);
    }   // for
 
    return (new Dimension((int) x, (int) y));
  } 

  /**
   * Zeichnet alle beteiligten Objekte. Jedes Objekt muss sich selber zeichnen koennen.
   */
  public synchronized void drawIt(Graphics2D g2) {
    int size = size();

    for (int i = 0; i < size; i++) {
      ((AnimationObject) elementAt(i)).paint(g2);
    }   // for
   }    // drawIt
 
  /**
   * Return true falls zwei Objekte zusammen getroffen sind.
   * 
   * @param r1 Object 1
   * @param r2 Object 2
   * 
   * @return true wenn die Objekte zusammen getroffen sind, andernfalls false.
   */
  private boolean collide(AnimationObject r1, AnimationObject r2) {
    double  distance = Geometrie.getDistance(r1.getX(), r1.getY(), r2.getX(), 
                                             r2.getY());

    if (distance < r1.getDiameter()) {
      return true;
    }   // if
 
    return false;
  }   // collide
 
  /**
   * Ein Objekt wird aus dem Clipboard in die Umgebung eingefuegt, dabei wird eine
   * freie Postionen ausgehend von (0,0) gesucht. Die Suche nach einer Luecke erfolgt
   * zuerst in der X-Richtung dann in der Y-Richtung, d.h. Zeilenweise wird abgetastet.
   */
  public boolean getClipboard(Simulator simulator, Umgebung umgebung) {
    boolean more = true;

    if (clipboardObject != null) {
      int             radius = clipboardObject.getDiameter() / 2;
      int             size = size();
      AnimationObject r2;
      int             y;
      int             x = radius;

      for (y = radius; y < umgebung.getHeight() - radius && more; y += radius) {
        for (x = radius; x < umgebung.getWidth() - radius && more; 
             x += radius) {
          clipboardObject.setPos(x, y);

          int j = 0;

          do {
            r2 = (AnimationObject) elementAt(j);
            j++;
          } while (j < size &&!collide(clipboardObject, r2));

          more = collide(clipboardObject, r2);
        }   // for
       }    // for
 
      if (!more) {
        setFreeIDNumber(clipboardObject);
        add(clipboardObject);
        setMark(simulator, size() - 1, true);

        if (clipboardObject instanceof Roboter) {
          // Es muss eine neue Kopie fuer das Reset bereitgestellt werden
          ((Roboter) clipboardObject).setMemento((Roboter) clipboardObject);
        }   // if

  

        try {

          // Neues Objekt fuer die naechste Kopie bereitstellen
          clipboardObject = ((AnimationObject) clipboardObject).cloneObject();
        } catch (Exception e) {
          e.printStackTrace();
        }   // catch
         finally {
          observable.notifyObservers(this);
        } 
      }     // if
     }      // if
 
    return !more;
  } 

  /**
   * Dateiverzeichnis welches die XML- und Propertie-Dateien enthaelt.
   * Diese Dateien beschreiben das Verhalten der Objekte. Wobei fuer die Roboter
   * XML verwendet wird.
   */
  public String getDirOfRobotData() {
    // Java Web Start findet diesen Pfad nicht
    // URL     url = ClassLoader.getSystemResource(DIR_OBJECTS);
    URL     url = ClassLoader.getSystemResource(DIR_OBJECTS);
    String  strUrl;
    if (url != null) {
        File  f = new File(url.getFile());
        if(!f.exists()) {
          // DIR_OBJECTS ist nur in der JAR-Datei enthalten
          url = null;
        }  // if
    } // if 
    if(url == null) {

      // Werden die Roboter nicht im classpath gefunden oder sind sie in einem Jar-File,
      // so werden diese im User-Home/DIR_OBJECTS gesucht.
      strUrl = System.getProperty("user.home") + "/" + DIR_OBJECTS;
      checkObjectDir(strUrl);
    } else {
      strUrl = url.getFile();
    } 

    /* System.out.println(strUrl); */

    return strUrl;
  }   // getDirOfRobotData
 
  /**
   * Rueckgabe der Dateiendung "xml" fuer Roboter-Beschreibung
   */
  public String getRobotSuffix() {
    return SUFFIX_XML;
  } 

  /**
   * Rueckgabe der Dateiendung "properties" fuer Aktoren.
   */
  public String getActorSuffix() {
    return SUFFIX_PROP;
  } 

  /**
   * Erstelle eine eindeutige ID fuer das Objekte. Dabei ist das erste
   * Zeichen die Art des Objektes und die folgenden Zeichen eine inkrementiert Zahl.
   * Das erste Zeichen und die inkrementierte Zahl identifizieren das Objekt eindeutig.
   * 
   * @param ao Das Objekt fuer welches die ID erweitert werden muss. Beim Eintritt
   * besteht die ID nur aus einem Zeichen welche die Art bezeichnet.
   */
  public void setFreeIDNumber(AnimationObject ao) {
    int       size = size();
    boolean[] isTakenNumber = new boolean[size + 1];
    String    letter = ao.getShortID().substring(0, 1);

    for (int i = 0; i < isTakenNumber.length; i++) {
      isTakenNumber[i] = false;
    } 

    for (int i = 0; i < size; i++) {

      // Ermittle alle vergebenen Nummern fuer eine ID
      String  id = ((AnimationObject) elementAt(i)).getShortID();

      if (id.startsWith(letter)) {
        int numberPart = Integer.parseInt(id.substring(1));

        if (numberPart < isTakenNumber.length) {

          // Streiche die Number aus dem Bereich der freien Nummern
          isTakenNumber[numberPart] = true;
        }   // for
       }    // if
     }      // for
 
    int i = 1;

    while (i < isTakenNumber.length && isTakenNumber[i]) {
      i++;
    }   // while
 
    ao.setShortID(letter + i);
  }   // getFreeObjektIdNumber
 
  /**
   * Loescht eine einzelnes Element im Vector.
   */
  public void removeOneElement(Simulator simulator, int workId) {
    if (workId >= 0) {
      setMark(simulator, workId, false);
      ((AnimationObject) this.elementAt(workId)).removeObject();
      super.removeElementAt(workId);
      observable.notifyObservers(this);
    }   // if
   } 

  /**
   * Loescht alle Elemente des Vectores.
   */
  public void removeAllElements(Simulator simulator) {
    int size = size();

    for (int i = 0; i < size; i++) {
      ((AnimationObject) this.elementAt(i)).removeObject();
    }   // for
 
    setMark(simulator, oldMarkId, false);
    super.removeAllElements();
    observable.notifyObservers(this);
  }   // removeAllElements
 
  /**
   * Damit der Simulator mit JWS bzw. mit Jar-Dateien gebraucht werden kann.
   * Mssen die in JAR-Dateien gehaltenen XML- und Property-Dateien exportiert werden.
   * Die XML werden unter DIR_OBJECT gespeichert.
   * Nur Dateien dir 
   *
   * @param dir Das Ziel-Verzeichnis der XML bzw. Property-Dateien
   *
   * @see
   */
  public void checkObjectDir(String dir) {
    File    f = new File(dir);
    String  strLine;

    if (!f.exists()) {
      f.mkdir();

      InputStream is = 
        Thread.currentThread().getContextClassLoader().getResourceAsStream(DIR_OBJECTS 
              + "/" + DIR_LIST);

      if (is != null) {
        try {
          BufferedReader  reader = 
            new BufferedReader(new InputStreamReader(is));
          // String          strUrl = System.getProperty("user.home");

          while ((strLine = reader.readLine()) != null) {
            strLine = strLine.trim();
            if(strLine.length() > 0) { 
              InputStream streamSource = 
                Thread.currentThread().getContextClassLoader().getResourceAsStream(DIR_OBJECTS 
                      + "/" + strLine);
              String      strDest = dir + "/" + strLine;
              copyFile(streamSource, strDest);
            }  
            
          }   // while
 
          reader.close();
        } catch (Exception e) {
          e.printStackTrace();
        }     // catch
       }      // if
     }        // if
   }          // checkObjectDir
 
  /**
   * Kopiert von einem InputStream auf eine Zieldatei.
   *
   * @param is InputStream
   * @param destFile Zieldatei
   */
  private void copyFile(InputStream fis, String destFile) {
    String  s;

    try {
      
      OutputStream fos = new FileOutputStream(destFile);
      byte buffer[] = new byte[0xffff];
      int nbytes;

      while(true) {
        if ((nbytes = fis.read(buffer)) == -1)
           break;
        fos.write(buffer, 0, nbytes);            
      } // while

      fis.close();
      fos.close();
    } catch (IOException ie) {
      ie.printStackTrace();
    } 
  } 

}

