package wesen.robot;

import java.awt.*;
import java.io.*;
import java.util.*;
import org.xml.sax.InputSource;
import org.exolab.castor.mapping.Mapping;
import org.exolab.castor.mapping.MappingException;
import org.exolab.castor.xml.Unmarshaller;
import javax.swing.JOptionPane;
import java.net.URL;

import org.exolab.castor.xml.ClassDescriptorResolver;
import org.exolab.castor.xml.Unmarshaller;
import org.exolab.castor.xml.util.ClassDescriptorResolverImpl;
import org.exolab.castor.xml.ValidationException;
import org.exolab.castor.xml.MarshalException;

import wesen.util.*;
import wesen.simulator.*;
import wesen.xml.*;


/**
 * Klasse welche einen einzelnen Roboter steuert und zeichnet.
 * 
 * @author Wesen-Team
 * @version 1.0
 */
public class Roboter extends AnimationObject{

  

  /**
   * Ueberschreiben der serialisierungs-ID, damit werden InvalidClassException ausgeschlossen.
   * Besser waere alle nicht fuer die Serialsierung benutzten Variablen als transient zu
   * deklarieren, dann wuerde bei einer Aenderung der nicht relevanten Variablen dieses
   * Problem nicht auftreten.
   */
  static final long           serialVersionUID = 5229886082550730613L;
  
  /**
   * Durchmesser des Kreises (Roboters)
   */
  public static final int     DIAMETER = 30;
  
  /**
   * Mappingdatei fuer das Einlesen der Robeter-XML Datei
   */
  public static final String  XML_MAPPING = "mapping.xml";

  /**
   * Fr Debuggingzwecke: Geschwindigkeiten der Motoren links und rechts anzeigen
   */ 
  public boolean debugValueSpeedsLeftRightTrend = false;


  /**
   * Fr Debuggingzwecke: Sichtwinkel der Sensoren anzeigen.
   */
  protected boolean debugSichtwinkel = false;
  
  /**
   * Fr Debuggingzwecke: Hilfswinkel zwischen Sensor und Aktor anzeigen.
   */
  protected boolean debugChiRight = false;
  
  /** 
   * Fr Debuggingzwecke: Distanz zwischen Sensor und Aktor anzeigen.
   * Range des Sensors anzeigen.
   */
  protected boolean debugValueDistanceRange = false;  


  protected Roboter memento;
  

  /**
   * Roboter Ressource Pfad
   */
  public static final String  XML_MAPPING_PATH = "resource";

  /**
   * Interne Geschwindigkeit eines Roboters, die von der Geschwindigkeit seiner
   * Motoren abhngig ist.
   * D.h. um xinc Pixel bewegt sich der Roboter bei einem Schritt (move) rechts.
   */
  private double              xinc;
  
  /**
   * Interne Geschwindigkeit eines Roboters, die von der Geschwindigkeit seiner
   * Motoren abhngig ist.
   * D.h. um yinc Pixel bewegt sich der Roboter bei einem Schritt (move) unten.
   */  
  private double              yinc;
  
  /**
   * Enthaelt die X-Koordinate des anderen Objektes mit welchen dieses Objekt zusammengestossen 
   * ist 
   */
  private double              coll_x;
  /**
   * Enthaelt die Y-Koordinate des anderen Objektes mit welchen dieses Objekt zusammengestossen 
   * ist 
   */
  private double              coll_y;
  
  /**
   * Bei true ist eine Kollision mit einem anderen Objekt
   */ 
  private boolean             collide;

  /**
   * Motor links
   * @supplierCardinality 1
   * @clientCardinality 1
   * @supplierQualifier hat links
   */
  public Motor                motorleft;

  /**
   * Motor rechts, public fuer XSD
   * @supplierCardinality 1
   * @clientCardinality 1
   * @supplierQualifier hat  rechts
   */
  public Motor                motorright;

  /**
   * Alle Sensoren dieses Roboters (rechts und links), public fuer XSD
   */
  public Vector               sensors;

  /**
   * Motor/Rad Abstand (entspricht Breite des Roboters)
   */
  double                      motorspace;

  /**
   * Zeitpunkt der letzten Bewegung
   */
  long                        lastmove;

  /**
   * Winkel des Roboters (Im Uhrzeigersinn in Radian).
   * Nach rechts schauend ist alpha=0.
   */
  double                      alpha = 0;  // -Math.PI/4;

  /**
   * Der origiale Winkel des Roboters (Im Uhrzeigersinn in Radian)
   */
  double                      originAlpha = alpha;

  /**
   * @link aggregationByValue
   * @clientCardinality 1
   * @supplierCardinality 1..*
   * @directed 
   */
  /*#  private SensorLicht lnkSensorLicht; */

  /**
   * Kreiert einen Roboter
   */
  public Roboter() {
  }

  /**
   * Kreiert einen Roboter welcher normalwerweise vom Benutzer interaktiv ins "Leben"
   * gerufen wurde.
   * 
   * @param originX Position in der X-Achse
   * @param originY Position in der Y-Achse
   * @param fileName Name der Datei welche die erweiterten Eigenschaften des Aktoren
   * enthaelt.
   * @param path Path wo sich die Datei befindet
   */
  public Roboter(int originX, int originY, String fileName, String path) {

    // setBaseValues(originX, originY, fileName, path);
  } 

  /**
   * Im ersten Simulationscyklus wird nur diese Methode durchlaufen
   */
  void initRoboter() {
    lastmove = System.currentTimeMillis();
  } 

  /**
   * Hiermit werden die Daten aus der XML-Datei eingelesen. 
   * Zuerst wird die XML-Datei gegen XSD validiert, danchach wir diese eingelsen.
   * 
   * @param originX Setzen des Roboters in der X-Achse
   * @param originY Setzen des Roboters in der Y-Achse
   * @param fileName Name der Datei
   * @param path Pfad wo sich die Datei befindet
   */
  static public Roboter getNewInstanceXML(int originX, int originY, 
                                          String fileName, String path) {
    Roboter robot = null;
    Mapping mapping = new Mapping();
    wesen.xml.Roboter tmpRoboter = null;
    
    try {
    
	    // Pruefen von XML gegen XSD. Diese Objekte werden dann nicht mehr benoetigt
      // Ziemlich viel Overhead.
	  tmpRoboter = wesen.xml.Roboter.unmarshal(new FileReader(path + "/" + fileName));


      // Mappingdaten vom Mappingfile laden
      
      /*
      String absolutName = ClassLoader.getSystemResource(XML_MAPPING_PATH).getFile();
      mapping.loadMapping(absolutName + "/" + XML_MAPPING);
      */
      
      URL absolutName = Thread.currentThread().getContextClassLoader().getResource(XML_MAPPING_PATH + "/" 
         + XML_MAPPING);
      mapping.loadMapping(absolutName);
      
      
      // Unmarshall XML nach Java Objekt
      Unmarshaller  unmar = new Unmarshaller(mapping);

      robot = (Roboter) unmar.unmarshal(new InputSource(new FileReader(path 
              + "/" + fileName)));

      /*
       * robot = (Roboter) unmar.unmarshal(getClass().getResourceAsStream("/" + path + "/"
       * + fileName));
       */
      robot.diameter = DIAMETER;
      robot.x = robot.originX = originX;
      robot.y = robot.originY = originY;
      robot.fileName = fileName;
     /*
      robot.color = ColorString.getColor(robot.strColor);
     */ 
      // Radabstand entspricht gerade dem Durchmesser
      robot.motorspace = robot.diameter;   
      // In der XML-Datei ist nur ein Motor definiert, daher wird der zweite 
      // Motor geklonnt.
      robot.motorright = robot.motorleft.cloneObject();
      
      robot.setSensorToRoboter(robot);

/*

      // Setze Sensoren (TODO: dynamisch)
      robot.sensors = new Vector();

      //Whle den Robotertyp 1, 2a, 2a, 3a, 3b
    

      // Typ 1 (Primitiv)
      // Info: Sensor -> Beide Sensoren je auf beide Motoren und +
      if (robot.shortID.startsWith("P")){
        System.out.println("Primitiv");
        robot.sensors.addElement(new SensorLicht(false, 100, robot, 't', 't'));      
        robot.sensors.addElement(new SensorLicht(true, 100, robot, 't', 't'));
      }

      // Typ 2a (ngstlich)
      // Info: Sensor -> Motor gerade und +
      if (robot.shortID.startsWith("T")){
        System.out.println("Aengstlich");
        robot.sensors.addElement(new SensorLicht(false, 100, robot, '+', '\0'));      
        robot.sensors.addElement(new SensorLicht(true, 100, robot, '\0', '+'));
      }
    
      
      // Typ 2b (Aggressiv)
      // Info: Sensor -> Motor bers Kreuz und +
      // Param: SensorLicht(isRight, range, robot, motorLinks, motorRechts)
      if (robot.shortID.startsWith("A")){
        System.out.println("Aggresiv");
        robot.sensors.addElement(new SensorLicht(false, 100, robot, '\0', '+'));      
        robot.sensors.addElement(new SensorLicht(true, 100, robot, '+', '\0'));
      }
      
    
      // Typ 3a (Lieb direkt)
      // Info: Sensor -> Motor bers Kreuz und -
      if (robot.shortID.startsWith("D")){
        System.out.println("Lieb direkt");
        robot.sensors.addElement(new SensorLicht(false, 100, robot, '\0', '-'));      
        robot.sensors.addElement(new SensorLicht(true, 100, robot, '-', '\0'));
      }      
    
      // Typ 3b (Lieb neugierig)
      // Info: Sensor -> Motor gerade und +
      if (robot.shortID.startsWith("N")){
        System.out.println("Lieb neugierig");
        robot.sensors.addElement(new SensorLicht(false, 100, robot, '-', '\0'));      
        robot.sensors.addElement(new SensorLicht(true, 100, robot, '\0', '-'));      
      }      
*/
     
    
  } catch(org.exolab.castor.xml.ValidationException ve) {
	    MessageHandler.showMessage(null, "XML-Datei: " + fileName + "|" 
	           + ve.getMessage(), "Fehler" , JOptionPane.ERROR_MESSAGE);
	}
	 catch(org.exolab.castor.xml.MarshalException me) {
	    MessageHandler.showMessage(null, "XML-Datei: " + fileName + "|" 
	           + me.getMessage(), "Fehler" , JOptionPane.ERROR_MESSAGE);

	}
	catch(MappingException me) {
	    me.printStackTrace();
	}    
	catch(IOException ioe) {
	    ioe.printStackTrace();
	}    
	catch(Exception e) {
	    e.printStackTrace();
	}    
  finally {
     return robot;
  }
   
  }  // getNewInstanceXML
 
  /**
   * Klonen dieses Objektes und rueckgabe der Referenz auf dieses neue Objekt.
   * 
   * @return Das erzeugte Objekt
   */
  public AnimationObject cloneObject() throws Exception {
    return (AnimationObject) (ObjectCloner.deepCopy(this));
  } 
  
  private void setSensorToRoboter(Roboter robot) {
      for (Enumeration el=robot.sensors.elements(); el.hasMoreElements(); ) {
         // Referenz von Sensor auf Roboter setzen und Position der Sensoren berechnen
         SensorLicht sl = (SensorLicht) el.nextElement();
         sl.setRoboter(robot);
         sl.calcNewPos();
      }
  }    



  /**
   * Update der Sensorwerte aufgrund der Aktoren.
   * Sollte nach jeder Roboterbewegung passieren.
   * Precondition: Die Positionen der Aktoren sind aktuell.
   * @param ov - Von diesem Vector werden die Aktorenobjekte gebraucht um die
   * @param Distanz zu den Sensoren und deren Werte zu berechnen.
   * @param index - Position dieses Roboters im ov
   */
  public void updateSensorAndMotorValues(ObjectVector ov, int index) {
    Sensor  sensor;

    // Loope ber alle Sensoren dieses Roboters
    for (int j = 0; j < sensors.size(); j++) {
      sensor = (Sensor) sensors.elementAt(j);

      ((SensorInterface) sensor).updateSensorValue(ov, index);
      sensor.setNewMotorSpeeds();
    }   // for j
   }    // method
 
  /**
   * Verschiebung des Roboters
   * @param ov ObjectVector der diese Methode aufruft
   * @param index dieses Roboters im ov
   */
  public synchronized void  move(ObjectVector ov, int index) {

    
    long  now = System.currentTimeMillis();
    long  dt = now - lastmove;  // Vergangene Zeit zur letzten Bewegung

    lastmove = now;   // Zeit speichern

    // Geschwindigkeit anpassen via Aktoren
    // Geschwindigkeit anpassen via Zeit
    motorleft.CalcNewDeltas(motorspace, dt, alpha);
    motorright.CalcNewDeltas(motorspace, dt, alpha);

    /* Passe xinc und yinc an Motorgeschwindigkeit an */
    xinc = motorleft.deltaX / 2 + motorright.deltaX / 2;
    yinc = motorleft.deltaY / 2 + motorright.deltaY / 2;

    // Neuer Winkel berechnen
    this.setNewAlpha();

    // Sensorpositionen neu berechnen
    Sensor.calcNewPosForMultipleSensors(this.sensors);

    // Sensorwerte updaten
    this.updateSensorAndMotorValues(ov, index);   // todo

    // double d=dt/100;
    // System.out.println("Delta Time:" + dt + " ,Seconds: " + d);

    /*
     * System.out.println("motorleft : x=" + motorleft.deltaX + ", y="
     * + motorleft.deltaY);
     * System.out.println("motorright: x=" + motorright.deltaX + ", y="
     * + motorright.deltaY);
     * System.out.println("New Alpha :" + alpha);
     * System.out.println();
     */
    if (collide) {
      double  xvect = coll_x - x;
      double  yvect = coll_y - y;

      if ((xinc > 0 && xvect > 0) || (xinc < 0 && xvect < 0)) {
        xinc = -xinc;
      } 

      if ((yinc > 0 && yvect > 0) || (yinc < 0 && yvect < 0)) {
        yinc = -yinc;
      } 

      collide = false;
    } 

    x += xinc;
    y += yinc;

    if (x < diameter / 2 
            || x > Simulator.getUmgebung().getWidth() - diameter / 2) {

      // Der Roboter ist in der X-Achse in eine Seitenwand gefahren
      xinc = -xinc;
      x += xinc;
    } 

    if (y < diameter / 2 
            || y > Simulator.getUmgebung().getHeight() - diameter / 2) {

      // Der Roboter ist in der Y-Achse in eine Seitenwand gefahren
      yinc = -yinc;
      y += yinc;
    } 
  } 

  /**
   * Berechnet den neuen Winkel Alpha
   */
  public void setNewAlpha() {
    alpha = this.alpha - motorleft.deltaAngle + motorright.deltaAngle;
  } 


  /**
   * Setzen des Roboter an eine neuen Position in der Umgebung.
   * Normalerweise geschieht diese durch Benutzerinteraktionen.
   * 
   * @param x Position in der X-Achse
   * @param y Position in der Y-Achse
   */
  public void setPos(int x, int y) {
    super.setPos(x, y);
    Sensor.calcNewPosForMultipleSensors(this.sensors);
  } 

  /**
   * Bei einem Zusammenstoss mit einem anderen Objekt
   * 
   * @param b Anderes Objekt
   * 
   * @see wesen.robot.ObjectVector#collide(AnimationObject, AnimationObject)
   */
  public void hit(AnimationObject b) {
    if (!collide) {
      coll_x = b.getX();
      coll_y = b.getY();
      collide = true;
    }   // if
   } 

  /**
   * Sich selber speichern. Wird fuer das Speichern der Welt benutzt.
   * 
   * @param oos OutputStram
   * 
   * @throws IOException Im Fehlerfall wird die Exception geworfen
   */
  public void saveAnimationObject(ObjectOutputStream oos) throws IOException {
    oos.writeObject(this.getClass());
    oos.writeDouble(x);
    oos.writeDouble(y);
    oos.writeUTF(fileName);
    oos.writeBoolean(debugValueSpeedsLeftRightTrend);
    oos.writeBoolean(debugSichtwinkel);
    oos.writeBoolean(debugChiRight);
    oos.writeBoolean(debugValueDistanceRange);
    oos.writeDouble(alpha);
  } 

  /**
   * Damit kann diese Debugoption abgefragt werden.
   */
  public void setDebugValueSpeedsLeftRightTrend(boolean debugValueSpeedsLeftRightTrend){
    this.debugValueSpeedsLeftRightTrend = debugValueSpeedsLeftRightTrend;
  }  

  /**
   * Damit kann diese Debugoption gesetzt werden.
   */
  public boolean getDebugValueSpeedsLeftRightTrend (){
    return debugValueSpeedsLeftRightTrend;
  }  



  /**
   * Einlesen der serailisierten Objekte dieses Objektes.
   * 
   * @param ois InputStream
   * @param path Pfad wo sich die Beschreibungsdatei befindet
   * 
   * @throws IOException Diese wird geworfen
   */
  public AnimationObject loadAnimationObject(ObjectInputStream ois, 
                                             String path) throws IOException {
    
    Roboter roboter = getNewInstanceXML((int) ois.readDouble(), 
                                        (int) ois.readDouble(), 
                                        (String) ois.readUTF(), path);
    // Die folgende Werte muessen immer gelesen werden, auch wenn der Roboter
    // nicht mehr instanziert werden kann.
    
    boolean vslr = ois.readBoolean();
    boolean dsw = ois.readBoolean();
    boolean dcr = ois.readBoolean();
    boolean dbdr = ois.readBoolean();
    double alpha = ois.readDouble(); 

    if(roboter != null) {
      roboter.debugValueSpeedsLeftRightTrend = vslr;
      roboter.debugSichtwinkel = dsw;
      roboter.debugChiRight = dcr;
      roboter.debugValueDistanceRange = dbdr;
      roboter.setOriginAlpha(alpha);
    }  

    return roboter;
  } 

  /**
   * Y Position des Roboters wird gesetzt
   */
  public void setY(double y) {
    this.y = y;
    Sensor.calcNewPosForMultipleSensors(this.sensors);
  } 

  /**
   * X Position des Roboters wird gesetzt
   */
  public void setX(double x) {
    this.x = x;
    Sensor.calcNewPosForMultipleSensors(this.sensors);
  } 
  
  /**
   * Zuesaetzlich muss noch der Winkel zurueckgesetzt werden.
   * 
   * @see wesen.robot.AnimationObject#resetPosition
   */
  public void resetPosition() {
    super.resetPosition();
    alpha = originAlpha;
    Sensor.calcNewPosForMultipleSensors(this.sensors);
  } 
  
  
  
  Roboter getMemento() {
    resetPosition();
    memento.setX(getX());
    memento.setY(getY());
    memento.setDebugChiRight(this.getDebugChiRight());
    memento.setDebugSichtwinkel(this.getDebugSichtwinkel());
    memento.setDebugValueDistanceRange(this.getDebugValueDistanceRange());
    memento.setDebugValueSpeedsLeftRightTrend(this.getDebugValueSpeedsLeftRightTrend());
    memento.setColor(this.getColor());
    memento.alpha = memento.originAlpha = this.originAlpha;
    memento.setSensorToRoboter(memento);
    
    memento.setMemento(memento);
    memento.beanEditor = this.beanEditor;
    if(memento.beanEditor != null) {
      memento.beanEditor.setObject(memento);
    }  
    return memento;
  }  
  
  void setMemento(Roboter rob) {
    try {
      memento = (Roboter) rob.cloneObject();
      
    } catch (Exception e) {
      e.printStackTrace();
    }
  }  
  

  /**
   * Damit kann diese Debugoption abgefragt werden.
   */
  public boolean getDebugSichtwinkel() {
    return debugSichtwinkel;
  } 
  
  /**
   * Damit kann diese Debugoption gesetzt werden.
   */
  public void setDebugSichtwinkel(boolean debugSichtwinkel) {
    this.debugSichtwinkel = debugSichtwinkel;
  }  

  /**
   * Damit kann diese Debugoption abgefragt werden.
   */
  public boolean getDebugChiRight() {
    return debugChiRight;
  }
  
  /**
   * Damit kann diese Debugoption gesetzt werden.
   */  
  public void setDebugChiRight(boolean debugChiRight) {
    this.debugChiRight = debugChiRight;
  }  

  /**
   * Damit kann diese Debugoption abgefragt werden.
   */  
  public boolean getDebugValueDistanceRange() {
    return debugValueDistanceRange;
  }  

  /**
   * Damit kann diese Debugoption gesetzt werden.
   */
  public void setDebugValueDistanceRange(boolean debugValueDistanceRange) {
    this.debugValueDistanceRange = debugValueDistanceRange;
  }  




  /**
   * Rueckgabe des Winkel in Radian
   * 
   * @return Rueckgabe des Winkel in Radian
   */
  public double getAlpha() {
    return alpha;
  } 

  /**
   * Setzen des Winkel in Radian
   * 
   * @param alpha Winkel in radian
   */
  public void setOriginAlpha(double alpha) {
    this.alpha = this.originAlpha = alpha;
    Sensor.calcNewPosForMultipleSensors(this.sensors);
  } 

  /**
   * Setzen des Winkel in Graden.
   * 
   * @param dGrad Winkel in Grad
   */
  public void setGrad(double dGrad) {
    setOriginAlpha(Math.toRadians(dGrad));
  } 

  /**
   * Zeichnet den Roboter
   * 
   * @param gr Grafikkontext
   * @see wesen.robot.AnimationObject#paint(Graphics)
   */
  public void paint(Graphics gr) {
    gr.setColor(Color.black);
    gr.fillOval(getDrawX(), getDrawY(), diameter, diameter);
    gr.setColor(color);

    // Zeichnet Sensoren
    // Sensor sensor=(Sensor)this.sensors.elementAt(0);
    // gr.drawLine((int)sensor.posX, (int)sensor.posY, (int)sensor.posX, (int)sensor.posY);
    paintSensors(gr);
    
    // Zeichnet den Roboter
    gr.fillOval(getDrawX() + 1, getDrawY() + 1, diameter - 2, diameter - 2);
    gr.setColor(Color.black);

    // Zeichnet seine Bewegungsrichtung
    gr.drawLine((int) x, (int) y, (int) (x + Math.cos(alpha) * diameter / 2), 
                (int) (y - Math.sin(alpha) * diameter / 2));  // (x1,y1,x2,y2)


    super.drawMark(gr);
  }                                                           // paint
 
  /**
   * Zeichnet die Sensoren dieses Roboters
   * (Den ersten linken und den ersten rechten)
   * @param gr Grafikkontext
   */
  private void paintSensors(Graphics gr) {
    boolean blnPaintedRight = false, blnPaintedLeft = false, blnPaint = false;

    for (int i = 0; i < sensors.size(); i++) {
      Sensor  sensor = (Sensor) sensors.elementAt(i);

      // Mache dass erster linker, resp. rechter Sensor gezeichnet wird
      if (sensor.blnIsRight &&!blnPaintedRight) {
        blnPaint = true;
        blnPaintedRight = true;
      } else if (!sensor.blnIsRight &&!blnPaintedLeft) {
        blnPaint = true;
        blnPaintedLeft = true;
      } 

      if (blnPaint) {

        // Zeichne den Sensor als Punkt
        gr.drawLine((int) sensor.posX, (int) sensor.posY, (int) sensor.posX, 
                    (int) sensor.posY);

        // Zeichne den Sensor als Linie
        gr.drawLine((int) x, (int) y, (int) sensor.posX, (int) sensor.posY);

        blnPaint = false;
      } 

      // Fertig?
      if (blnPaintedRight && blnPaintedLeft) {
        break;
      } // if
    }   // for i
   }    // paintSensors

}

